import glob
import os
import re
from collections import defaultdict
from subprocess import check_output, CalledProcessError

from jinja2 import Environment, PackageLoader

from s3pypi import __prog__
from s3pypi.exceptions import S3PyPiError

__author__ = 'Matteo De Wint'
__copyright__ = 'Copyright 2016, November Five'
__license__ = 'MIT'


class Package(object):
    """Python package."""

    def __init__(self, name, files):
        self.name, self.version = name.rsplit('-', 1)
        self.files = set(files)

    def __str__(self):
        return '%s-%s' % (self.name, self.version)

    def _attrs(self):
        return self.name, self.version

    def __lt__(self, other):
        return self.version < other.version

    def __eq__(self, other):
        return isinstance(other, Package) and self._attrs() == other._attrs()

    def __hash__(self):
        return hash(self._attrs())

    @property
    def directory(self):
        return re.sub('[-_.]+', '-', self.name)

    @staticmethod
    def _find_package_name(text):
        match = re.search('^(copying files to|making hard links in) (.+)\.\.\.', text, flags=re.MULTILINE)

        if not match:
            raise RuntimeError('Package name not found in:\n' + text)

        return match.group(2)

    @staticmethod
    def _find_wheel_name(text):
        match = re.search('creating \'.*(dist.*\.whl)\' and adding', text, flags=re.MULTILINE)

        if not match:
            raise RuntimeError('Wheel name not found in:\n' + text)

        return match.group(1)

    @staticmethod
    def create(wheel=True, sdist=True):
        cmd = ['python', 'setup.py', 'sdist', '--formats', 'gztar']

        if wheel:
            cmd.append('bdist_wheel')

        try:
            stdout = check_output(cmd).decode().strip()
        except CalledProcessError as e:
            raise RuntimeError(e.output.rstrip())

        name = Package._find_package_name(stdout)
        files = []

        if sdist:
            files.append(name + '.tar.gz')

        if wheel:
            files.append(os.path.basename(Package._find_wheel_name(stdout)))

        return Package(name, files)


class Index(object):
    """Index of package versions, to be rendered to HTML."""

    template = Environment(loader=PackageLoader(__prog__, 'templates')).get_template('index.html.j2')

    def __init__(self, packages):
        self.packages = set(packages)

    @staticmethod
    def parse(html):
        filenames = defaultdict(set)

        for match in re.findall('<a href="((.+?-\d+\.\d+\.\d+).+)">', html):
            filenames[match[1]].add(match[0])

        return Index(Package(name, files) for name, files in filenames.items())

    def to_html(self):
        return self.template.render({'packages': self.packages})

    def add_package(self, package, force=False):
        if force:
            self.packages.discard(package)
        elif any(p.version == package.version for p in self.packages):
            raise S3PyPiError('%s already exists! You should use a different version.' % package)

        self.packages.add(package)



import os

import boto3
from botocore.exceptions import ClientError

from s3pypi.package import Index

__author__ = 'Matteo De Wint'
__copyright__ = 'Copyright 2016, November Five'
__license__ = 'MIT'


class S3Storage(object):
    """Abstraction for storing package archives and index files in an S3 bucket."""

    def __init__(self, bucket, secret=None, region=None, bare=False, private=False, profile=None):
        if profile:
            boto3.setup_default_session(profile_name=profile)        
        self.s3 = boto3.resource('s3', region_name=region)
        self.bucket = bucket
        self.secret = secret
        self.index = '' if bare else 'index.html'
        self.acl = 'private' if private else 'public-read'

    def _object(self, package, filename):
        path = '%s/%s' % (package.directory, filename)
        return self.s3.Object(self.bucket, '%s/%s' % (self.secret, path) if self.secret else path)

    def get_index(self, package):
        try:
            html = self._object(package, self.index).get()['Body'].read().decode('utf-8')
            return Index.parse(html)
        except ClientError:
            return Index([])

    def put_index(self, package, index):
        self._object(package, self.index).put(
            Body=index.to_html(),
            ContentType='text/html',
            CacheControl='public, must-revalidate, proxy-revalidate, max-age=0',
            ACL=self.acl
        )

    def put_package(self, package):
        for filename in package.files:
            with open(os.path.join('dist', filename), mode='rb') as f:
                self._object(package, filename).put(
                    Body=f,
                    ContentType='application/x-gzip',
                    ACL=self.acl
                )



from __future__ import print_function

import argparse
import sys

from s3pypi import __prog__
from s3pypi.exceptions import S3PyPiError
from s3pypi.package import Package
from s3pypi.storage import S3Storage

__author__ = 'Matteo De Wint'
__copyright__ = 'Copyright 2016, November Five'
__license__ = 'MIT'


def create_and_upload_package(args):
    package = Package.create(args.wheel, args.sdist)
    storage = S3Storage(args.bucket, args.secret, args.region, args.bare, args.private, args.profile)

    index = storage.get_index(package)
    index.add_package(package, args.force)

    storage.put_package(package)
    storage.put_index(package, index)


def parse_args(raw_args):
    p = argparse.ArgumentParser(prog=__prog__)
    p.add_argument('--bucket', required=True, help='S3 bucket')
    p.add_argument('--secret', help='S3 secret')
    p.add_argument('--region', help='AWS region')
    p.add_argument('--profile', help='AWS profile')
    p.add_argument('--force', action='store_true', help='Overwrite existing packages')
    p.add_argument('--no-wheel', dest='wheel', action='store_false', help='Skip wheel distribution')
    p.add_argument('--no-sdist', dest='sdist', action='store_false', help='Skip sdist distribution')
    p.add_argument('--bare', action='store_true', help='Store index as bare package name')
    p.add_argument('--private', action='store_true', help='Store S3 Keys as private objects')
    return p.parse_args(raw_args)


def main():
    args = parse_args(sys.argv[1:])

    try:
        create_and_upload_package(args)
    except S3PyPiError as e:
        print('error: %s' % e)
        sys.exit(1)


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

from s3pypi import __prog__, __version__


setup(
    name=__prog__,
    version=__version__,

    description='CLI tool for creating a Python Package Repository in an S3 bucket',

    author='Ruben Van den Bossche, Matteo De Wint',
    author_email='ruben@novemberfive.co, matteo@novemberfive.co',
    url='https://github.com/novemberfiveco/s3pypi',
    download_url='https://github.com/novemberfiveco/s3pypi/tarball/' + __version__,

    packages=find_packages(),
    package_data={__prog__: ['templates/*.j2']},

    install_requires=['boto3', 'Jinja2', 'wheel'],
    entry_points={'console_scripts': ['{0}={0}.__main__:main'.format(__prog__)]},
)
